﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace JeuAvecDesKilometres
{
    public delegate void EvenementCarte(Joueur LeJoueur,Carte LaCarte, Interdictions Raison); //signature de l'évènement, je déclare ce délégué en dehors de la classe, ainsi, il est visible dans tout l'espace de nom

    public class Joueur:INotifyPropertyChanged
    {
        /// <summary>
        /// Chronomètre pour le temps de jeu d'un coup fourré à partir du niveau "normal"
        /// </summary>
        /// <remarks>Ce chrono est une variable de classe, donc n'importe quel joueur peut le déclencher, et n'importe joueur peut le stopper</remarks>
        private static Stopwatch chrono = new Stopwatch();

        /// <summary>
        /// Quand un joueur d'un niveau à partir du niveau "normal" a la possibilité de jouer un coup fourré, il est stocké dans cette varible static
        /// </summary>
        private static Joueur coupFourrePossible;

        public event EvenementCarte CarteInterdite; //évènement signalant que le joueur tente de poser une carte à l'encontre des règles.

        public delegate void EvenementKilometres(Joueur LeJoueur, int Kilometres);//signature de l'évènement
        public event EvenementKilometres JaiGagne;//évènement signalant que la distance est atteinte
        public event EvenementKilometres DistanceDepassee;//évènement signalant que la distance est partagée

        public delegate void EvenementEmbuche(Joueur LeJoueur);//signature de l'évènement
        public event EvenementEmbuche EmbucheRecue;//évènement signalant une embuche reçue
        public event EvenementEmbuche CoupFourre;//évènement signalant un coup fourré
        public event EvenementEmbuche FiniMonTour;//évènement signalant que le joueur à fini son tour.
        public event EvenementEmbuche PiocherDabord;//évènement signalant que le joueur doit piocher avant de jouer.

        /// <summary>
        /// Constructeur 1
        /// </summary>
        /// <param name="LeNom">Nom du joueur</param>
        public Joueur(string LeNom):this(LeNom,Difficulte.Facile)//on appelle le constructeur plus complet avec une valeur par défaut
        {
        }

        /// <summary>
        /// Constructeur 2
        /// </summary>
        /// <param name="LeNom">Nom du joueur</param>
        /// <param name="LeNiveau">et son niveau</param>
        public Joueur(string LeNom, Difficulte LeNiveau)
        {
            Nom = LeNom;
            Main = new ObservableCollection<Carte>();
            Jeu = new ObservableCollection<Carte>();//initialisation de la collection
            Jeu.CollectionChanged += Jeu_CollectionChanged;//Abonnement à l'évènement signalant un changement
            Niveau = LeNiveau;
            Score = new Dictionary<int, int>();

            NouvellePartie();
        }

        #region Propriétés

        private string nom;
        /// <summary>
        /// Nom du joueur
        /// </summary>
        public string Nom
        {
            get { return nom; }
            private set
            {
                if (nom != value)
                {
                    nom = value;
                    RaisePropertyChanged("Nom");
                }
            }
        }

        private ObservableCollection<Carte> main;
        /// <summary>
        /// Collection de cartes dans la main du joueur
        /// </summary>
        public ObservableCollection<Carte> Main
        {
            get { return main; }
            set
            {
                if (main != value)
                {
                    main = value;
                }
            }
        }

        /// <summary>
        /// Collection de cartes representant le jeu du joueur
        /// </summary>
        public ObservableCollection<Carte> Jeu { get; set; }

         private Difficulte niveau;
       /// <summary>
        /// Niveau de difficulté du joueur
        /// </summary>
        public Difficulte Niveau
        {
            get { return niveau; }
            set
            {
                if (niveau != value)
                {
                    niveau = value;
                    RaisePropertyChanged("Niveau");
                }
            }
        }

        private Carte derniereCarte;
        /// <summary>
        /// Dernière carte posée dans le jeu, y compris les embuches
        /// </summary>
        public Carte DerniereCarte
        {
            get { return derniereCarte; }
            private set
            {
                if (derniereCarte != value)
                {
                    derniereCarte = value;
                    RaisePropertyChanged("DerniereCarte");
                }
            }
        }

        /// <summary>
        /// Propriété en lecture seule, dont la valeur est le résultat d'un calcul
        /// </summary>
        /// <remarks>Etant en lecture seule, il faudra signaler le changement de valeur potentiel ailleurs dans cette classe</remarks>
        public int Points
        {
            get
            {
                int lesPoints = 0;
                foreach (Carte laCarte in Jeu)
                {
                    lesPoints += laCarte.Points;//grâce au polymorphisme, on peut additionner les points, peu importe le type de carte.
                }


                int bonus = 0;

                if (!Moteur.PartieEnCours)
                {
                    //à la fin de la partie si un (ou plusieurs) joueur n'a pas pu poser de kilomètre, les autres marquent 500 points de bonus
                    foreach (Joueur leJoueur in Moteur.Joueurs)
                    {
                        if (leJoueur == this)
                            continue;//on passe à l'itération suivante

                        if (leJoueur.DistanceParcourue == 0)
                            bonus += 500;//on ajoute 500 au bonus pour se joueur sans distance
                    }
                }

                if (laDistanceParcourue == 1000)//bonus du gagnant
                {
                    bonus += 400;//400 pour avoir gagné

                    if (Jeu.OfType<Kilometre>().Count(k => k.Distance == 200) == 0)//pour la clause if, je filtre la collection Main, sur le type Kilometre, et je compte le nombre de Kilometre dont la distance vaut 0
                        bonus += 200;// 200 de plus si aucune carte 200 utilisée
                }
                    

                return lesPoints + bonus;
            }
        }

        private Dictionary<int,int> score;
        /// <summary>
        /// Dictionnaire stockant pour chaque partie les points réalisés
        /// </summary>
        public Dictionary<int,int> Score
        {
            get { return score; }
            private set
            {
                if (score != value)
                {
                    score = value;
                    RaisePropertyChanged("Score");
                }
            }
        }

        int laDistanceParcourue = 0;
        /// <summary>
        /// Distance parcourrue pour cette partie
        /// </summary>
        /// <remarks>Cette propiété nécessite une variable interne car selon le niveau la distance est affichée ou pas, mais on doit connaitre la distance pour savoir si le joueur gagne ou dépasse
        /// Etant en lecture seule, il faudra signaler le changement de valeur potentiel ailleurs dans cette classe</remarks>
        public int DistanceParcourue
        {
            get
            {
                switch (Niveau)
                {
                    case Difficulte.Facile:
                    case Difficulte.Normale:
                        return laDistanceParcourue;

                    default:
                        return 0;
                }
            }
        }

        private TypeEmbuche embucheEnCours;
        /// <summary>
        /// Embûches que subit actuellement le joueur
        /// </summary>
        public TypeEmbuche EmbucheEnCours
        {
            get { return embucheEnCours; }
            set
            {
                if (embucheEnCours != value)
                {
                    embucheEnCours = value;
                    RaisePropertyChanged("EmbucheEnCours");
                }
            }
        }


        private TypeEmbuche immunite;
        /// <summary>
        /// Immunités dont dispose le joueur
        /// </summary>
        public TypeEmbuche Immunite
        {
            get { return immunite; }
            set
            {
                if (immunite != value)
                {
                    immunite = value;
                    RaisePropertyChanged("Immunite");
                }
            }
        }

        private bool actif;
        /// <summary>
        /// Retourne ou définit s'il s'agit du joueur dont c'est le tour
        /// </summary>
        public bool Actif
        {
            get { return actif; }
            set
            {
                if (actif != value)
                {
                    actif = value;
                    RaisePropertyChanged("Actif");
                }
            }
        }
        
        #endregion

        #region Méthodes
        /// <summary>
        /// (Re)initialise les variables
        /// </summary>
        public void NouvellePartie()
        {
            EmbucheEnCours = TypeEmbuche.FeuRouge;
            Immunite = TypeEmbuche.Aucune;
            Main.Clear();
            Jeu.Clear();
        }

        /// <summary>
        /// Signale au joueur que la partie est finie, il stocke ses points dans le dictionnaire.
        /// </summary>
        /// <param name="Numero"></param>
        public void FinPartie(int Numero)
        {
            Score.Add(Numero, Points);
        }

        /// <summary>
        /// Traitements à faire après qu'une carte ait été ajoutée au jeu
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Jeu_CollectionChanged(object sender, System.Collections.Specialized.NotifyCollectionChangedEventArgs e)
        {
            if (e.Action != System.Collections.Specialized.NotifyCollectionChangedAction.Add)
                return;

            coupFourrePossible = null;//RAZ de la variable, si un joueur joue avant le coup fourré, celui qui pouvait perd cette opportunité

            DerniereCarte = Jeu[e.NewStartingIndex];

            if (DerniereCarte is Embuche)//si la dernière carte recue est une embuche, car Jeu contient des Cartes
            {
                if (this.EmbucheRecue != null)//on vérifie que "quelqu'un" soit abonné à l'évènement
                    this.EmbucheRecue(this);//on émet l'évènement

                Embuche lEmbuche = (Embuche)DerniereCarte;//pour travailler sur une Embuche, à partir d'une Carte, il faut "caster" (forcer la conversion) la Carte en Embuche

                //on regarde si le joueur dispose de la botte pour faire un coup fourré
                foreach (Carte laCarte in Main)
                {
                    if (laCarte is Botte && ((Botte)laCarte).EmbucheImmunisee.HasFlag(lEmbuche.Type))
                    {

                        if (Niveau == Difficulte.Facile)
                        {
                            Joue((Botte)laCarte);//On joue la botte
                            FaitCoupFourre((Botte)laCarte);//on joue le coup fourré
                            return;//on sort de la méthode, ça n'est pas la peine de continuer à cherche ce que l'on a déjà trouvé
                        }
                        else
                        {
                            //on laisse au joueur 10 secondes pour réagir, on lance donc le chrono
                            coupFourrePossible = this;
                            chrono.Start();
                        }
                    }
                }

                EmbucheEnCours |= lEmbuche.Type;//on affecte l'embûche
                if (lEmbuche.Type != TypeEmbuche.LimitationVitesse && !Immunite.HasFlag(TypeEmbuche.FeuRouge))
                    EmbucheEnCours |= TypeEmbuche.FeuRouge;//on affecte un feu rouge pour que le joueur soit obligé de mettre un feu vert

            }
            else if (DerniereCarte is Kilometre)//si la dernière carte est un kilomètre, on calcule la distance totale.
            {
                Kilometre km = (Kilometre)DerniereCarte;
                int distance = km + laDistanceParcourue;

                if (distance > Moteur.DistanceAParcourir)
                {
                    if (this.CarteInterdite != null)
                        this.CarteInterdite(this, DerniereCarte, Interdictions.DistanceDepassee);

                    if (this.DistanceDepassee != null)
                        this.DistanceDepassee(this, distance);

                    return;
                }

                laDistanceParcourue = distance;//on valide la distance
                if (laDistanceParcourue == Moteur.DistanceAParcourir)
                {
                    if (this.JaiGagne != null)
                        this.JaiGagne(this, laDistanceParcourue);//on émet l'évènement avec le paramètre correspondant à la signature.

                }
                RaisePropertyChanged("DistanceParcourue");//signale un changement de distance
            }
            RaisePropertyChanged("Points");//signale un éventuel changement de points
        }

        /// <summary>
        /// Joue un coup fourré
        /// </summary>
        /// <param name="LaBotte"></param>
        private void FaitCoupFourre(Botte LaBotte)
        {
            LaBotte.CoupFourreParLeJoueur();//appelle la méthode qui va mettre le bonus de points
            Jeu.Remove(DerniereCarte);//on enlève l'embuche du jeu, ce qui va regénérer l'évènement Jeu_CollectionChanged, d'ou le premier test
            Moteur.Defausse.Add(DerniereCarte);//on jette l'embuche à la défausse

            //on signale que l'on joue le coup fourré
            if (this.CoupFourre != null)
                this.CoupFourre(this);

        }

        /// <summary>
        /// Méthode permettant de jouer une Carte dans son jeu
        /// </summary>
        /// <param name="LaCarte"></param>
        /// <remarks>Si cette méthode était publique, on se jouerait une embuche à soi-même</remarks>
        private void Joue(Carte LaCarte)
        {
            if (!AutorisationDeJouer(LaCarte))
                return;

            if (LaCarte is Kilometre)
            {
                switch (EmbucheEnCours)
                {
                    case TypeEmbuche.Aucune://on peut rouler
                        break;

                    case TypeEmbuche.LimitationVitesse://il n'y a aucune autre embûche
                        if (((Kilometre)LaCarte).Distance > 50)
                        {
                            AnnonceCarteInterdite(LaCarte, Interdictions.Limitation);
                            return;//on ne peut pas dépasser 50
                        }
                        break;

                    default://s'il y a n'importe quelle autre embuche, on ne peut pas rouler
                        AnnonceCarteInterdite(LaCarte, Interdictions.AttaqueEnCours);
                        return;
                }

                if (((Kilometre)LaCarte).Distance == 200 && Jeu.OfType<Kilometre>().Count(k => k.Distance == 200) == 2)
                {
                    //On ne peut pas joueur plus de 2 "200 km"
                    //à noter pour cette recherche, au lieu d'imriquer des boucles (comme plus haut), j'ai utilisé une requête Linq, qui filtre sur le type puis compte le nombre de "200 km"
                    AnnonceCarteInterdite(LaCarte, Interdictions.Max200);
                    return;
                }
            }
            else if (LaCarte is Parade)
            {
                Parade maParade = (Parade)LaCarte;
                if (!EmbucheEnCours.HasFlag(maParade.EmbucheParee))
                {
                    AnnonceCarteInterdite(LaCarte, Interdictions.ParadeIncoherente);
                    return;//on ne peut pas parer une embuche que l'on ne subit pas
                }

                if (maParade.EmbucheParee == TypeEmbuche.FeuRouge && !(EmbucheEnCours == TypeEmbuche.FeuRouge || EmbucheEnCours == (TypeEmbuche.FeuRouge | TypeEmbuche.LimitationVitesse)))
                {
                    AnnonceCarteInterdite(LaCarte, Interdictions.ParadeAvantFeuVert);
                    return;//on ne peut poser un feu vert qui si on est en "état" de feu rouge, ou feu rouge et limitation, toutes les autres embuches doivent avoir été parées avant
                }

                EmbucheEnCours ^= maParade.EmbucheParee;
            }
            else
            {
                //c'est forcément une botte, elle peut se poser n'importe quand même sans parer une embûche
                Botte laBotte = ((Botte)LaCarte);

                if (coupFourrePossible != null && coupFourrePossible == this)//le joueur doit avoir été attaqué, disposer de la botte et (réagir sous 10 s vérifié dans AutorisationDeJouer) et avant un autre joueur
                    FaitCoupFourre(laBotte);

                Immunite |= laBotte.EmbucheImmunisee;//immunise

                if (laBotte.EmbucheImmunisee.HasFlag(TypeEmbuche.FeuRouge))
                {
                    if (EmbucheEnCours.HasFlag(TypeEmbuche.FeuRouge))
                        EmbucheEnCours ^= TypeEmbuche.FeuRouge;//pare si besoin

                    if (EmbucheEnCours.HasFlag(TypeEmbuche.LimitationVitesse))
                        EmbucheEnCours ^= TypeEmbuche.LimitationVitesse;//pare si besoin
                }
                else if (EmbucheEnCours.HasFlag(laBotte.EmbucheImmunisee))
                    EmbucheEnCours ^= laBotte.EmbucheImmunisee;//pare si besoin
            }

            chrono.Reset();//Raz des variables
            coupFourrePossible = null;

            Main.Remove(LaCarte);//On enlève la carte de la main

            Carte superieure = Jeu.FirstOrDefault(c => c.Ordre > LaCarte.Ordre);//on cherche la première carte dont l'ordre est supérieur à la carte jouée.
            if (superieure != null)
                Jeu.Insert(Jeu.IndexOf(superieure), LaCarte);//si la carte supérieure existe, on insère la carte jouée à l'index de la carte supérieure
            else
                Jeu.Add(LaCarte);//sinon on ajoute la carte au jeu (donc à la fin)

            if (LaCarte is Botte)
                PiocheAutomatique();//si c'est une botte, en coup fourré ou non, le joueur rejoue
            else
                AnnonceFinTour();//Annonce que le joueur à fini de joueur, ce qui entraine que le "tour" passe au suivant qui pioche automatiquement

        }

        /// <summary>
        /// Méthode permettant de jouer un kilomètre
        /// </summary>
        /// <param name="Km"></param>
        /// <remarks>Je crée 3 surcharges pour être sûr qu'une embuche ne sera pas transmise en paramètre</remarks>
        public void Jouer(Kilometre Km)
        {
            Joue(Km);
        }

        /// <summary>
        /// Surcharge permettant de jouer une parade
        /// </summary>
        /// <param name="LaParade"></param>
        public void Jouer(Parade LaParade)
        {
            Joue(LaParade);
        }

        /// <summary>
        /// Surcharge permettant de jouer une botte
        /// </summary>
        /// <param name="LaParade"></param>
        public void Jouer(Botte LaBotte)
        {
            Joue(LaBotte);
        }

        /// <summary>
        /// Surcharge permettant de mettre une embuche à un adversaire
        /// </summary>
        /// <param name="Lembuche"></param>
        /// <param name="Cible"></param>
        public void Jouer(Embuche Lembuche, Joueur Cible)
        {
            if (!AutorisationDeJouer(Lembuche))
                return;

            if (Cible == this)//si le joueur cible est le joueur en cours, on génère une erreur
                throw new Exception("Le joueur ne peut pas se mettre une embuche à lui-même");

            if (Cible.Immunite.HasFlag(Lembuche.Type))//on ne peut pas poser ni une embûche  contre laquelle la cible est immunisée,
                AnnonceCarteInterdite(Lembuche, Interdictions.Immunite);
            
            else if (Lembuche.Type == TypeEmbuche.LimitationVitesse && Cible.EmbucheEnCours.HasFlag(TypeEmbuche.LimitationVitesse))//ni une limitation si elle y est déjà
                    AnnonceCarteInterdite(Lembuche, Interdictions.MaxLimitation);

            else if (Lembuche.Type != TypeEmbuche.LimitationVitesse && Cible.EmbucheEnCours != TypeEmbuche.Aucune && Cible.EmbucheEnCours != TypeEmbuche.LimitationVitesse)//ni les autres embûches s'il y a en déjà une (sauf la limitation)
                AnnonceCarteInterdite(Lembuche, Interdictions.DejaAttaque);

            else
            {
                Main.Remove(Lembuche);//On enlève la carte de la main

                Carte superieure = Cible.Jeu.FirstOrDefault(c => c.Ordre > Lembuche.Ordre);//on cherche la première carte dont l'ordre est supérieur à la carte jouée.
                if (superieure != null)
                    Cible.Jeu.Insert(Cible.Jeu.IndexOf(superieure), Lembuche);//si la carte supérieure existe, on insère la carte jouée à l'index de la carte supérieure
                else
                    Cible.Jeu.Add(Lembuche);//sinon on ajoute la carte au jeu (donc à la fin)

                AnnonceFinTour();
            }
        }

        /// <summary>
        /// Génère l'évènement
        /// </summary>
        private void AnnonceCarteInterdite(Carte LaCarte, Interdictions Raison)
        {
            if (this.CarteInterdite != null)
                this.CarteInterdite(this, LaCarte, Raison);
        }

        /// <summary>
        /// Pioche de façon automatique si le joueur a un niveau "bas"
        /// </summary>
        public void PiocheAutomatique()
        {
            if (Niveau == Difficulte.Facile)
                Pioche();
        }

        /// <summary>
        /// Pioche une carte
        /// </summary>
        public void Pioche()
        {
            if (Moteur.PartieEnCours && Moteur.JoueurEnCours != this)
            {
                AnnonceCarteInterdite(null, Interdictions.PasSonTour);
                return;
            }


            Carte cPiochee = Moteur.Pioche[0];
            Moteur.Pioche.Remove(cPiochee);//pioche la première carte

            if (Main.Count > 6)
            {
                AnnonceCarteInterdite(cPiochee, Interdictions.PiocheInterdite);
                return;
            }

            Carte superieure = Main.FirstOrDefault(c => c.Ordre > cPiochee.Ordre);//on cherche la première carte dont l'ordre est supérieur à la carte jouée.
            if (superieure != null)
                Main.Insert(Main.IndexOf(superieure), cPiochee);//si la carte supérieure existe, on insère la carte jouée à l'index de la carte supérieure
            else
                Main.Add(cPiochee);//sinon on ajoute la carte au jeu (donc à la fin) 
        }

        /// <summary>
        /// Teste si le joueur a le droit de jouer
        /// </summary>
        /// <param name="LaCarte"></param>
        /// <returns></returns>
        private bool AutorisationDeJouer(Carte LaCarte)
        {
            if (coupFourrePossible != null && chrono.ElapsedMilliseconds < 11e3)
            {
                if(coupFourrePossible != this && coupFourrePossible.Niveau == Difficulte.Normale)//si un joueur de niveau Normal a la possibilité de joueur un coup fourré on lui laisse 10s
                {
                    AnnonceCarteInterdite(LaCarte, Interdictions.Patienter);
                    return false;
                }
                if (coupFourrePossible == this && LaCarte is Botte && ((Embuche)coupFourrePossible.DerniereCarte).Type.HasFlag(((Botte)LaCarte).EmbucheImmunisee))//si un joueur joue son coup fourré
                {
                    return true;
                }
            }
            if(Moteur.JoueurEnCours != this)
            {
                 AnnonceCarteInterdite(LaCarte, Interdictions.PasSonTour);
                 return false;
            }
            if (Main.Count < 7 && this.PiocherDabord != null)
            {
                this.PiocherDabord(this);
                return false;
            }

            return true;
        }

        /// <summary>
        /// Défausse une carte
        /// </summary>
        /// <param name="LaCarte">Carte à défausser</param>
        /// <param name="PasseLeTour">Indique si le joueur doit passer son tour</param>
        public void Defausse(Carte LaCarte, bool PasseLeTour = true)
        {
            if (!AutorisationDeJouer(LaCarte))
                return;

            Moteur.Defausse.Add(LaCarte);
            Main.Remove(LaCarte);
            if (PasseLeTour)
                AnnonceFinTour();

        }

        /// <summary>
        /// Annonce que le tour est terminé
        /// </summary>
        private void AnnonceFinTour()
        {
            if (this.FiniMonTour != null)
                this.FiniMonTour(this);
        }

        /// <summary>
        /// Réécriture de ToString qui affiche le nom
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return Nom;
        }
        #endregion

        #region INotifyPropertyChanged
        public event PropertyChangedEventHandler PropertyChanged;
        /// <summary>
        /// Vérifie s'il y a un abonné à l'évènenemet et le génère
        /// </summary>
        /// <param name="Propriete"></param>
        private void RaisePropertyChanged(string Propriete)
        {
            if (this.PropertyChanged != null) PropertyChanged(this, new PropertyChangedEventArgs(Propriete));
        }

        #endregion

    }
}
